@@ -1,3 +1,4 @@
+/* XX: turned all  /stderr/ to /stdout/ 940906/mea */
 #include "hostenv.h"
 #include "mailer.h"
 #include <ctype.h>
@@ -1222,7 +1223,7 @@ regnext(p)
 STATIC char *regprop();
 
 /*
- - regdump - dump a regexp onto stdout in vaguely comprehensible form
+ - regdump - dump a regexp onto stderr in vaguely comprehensible form
  */
 void
 regdump(r)
