@@ -322,9 +322,10 @@ mkDate(isresent, unixtime)
 	long unixtime;
 {
 	register struct header	*h;
-	u_char *cp;
+	u_char *cp, *s;
 	static struct headerinfo *dateDesc;
 	
+	/* There is really only one kind of dateDesc.. */
 	if (dateDesc == NULL) {
 		dateDesc = find_header(spt_headers,
 				       isresent ? "resent-date" : "date");
@@ -338,6 +339,8 @@ mkDate(isresent, unixtime)
 	h->h_next = 0;
 	h->h_contents.d = unixtime;
 	cp = rfc822date(&(h->h_contents.d));
+	s = cp + strlen(cp) -1;
+	if (*cp != 0 && *s == '\n') *s = 0;
 	h->h_lines = makeToken(cp, strlen((char *)cp));
 	h->h_lines->t_type = Line;
 	/* OPTIMIZE: save this in a hashtable somewhere */
@@ -359,7 +362,7 @@ hdr_print(h, fp)
 	struct token *t;
 	struct address *ap;
 	struct addr *pp;
-	char *cp;
+	char *cp, *s;
 
 	if (h == NULL)
 		return;
@@ -417,7 +420,13 @@ hdr_print(h, fp)
 			fprintf(fp, " for ");
 			printAddress(fp, ap->a_tokens, 0);
 		}
-		fprintf(fp, "; %s", rfc822date(&(h->h_contents.r->r_time)));
+		putc(';', fp);
+		putc(' ', fp);
+		cp = rfc822date(&(h->h_contents.r->r_time));
+		s = cp + strlen(cp) -1;
+		if (*cp != 0 && *s == '\n') *s = 0;
+		fputs(cp, fp);
+		putc('\n', fp);
 		break;
 	case nilHeaderSemantics:
 		for (t = h->h_lines; t != NULL; t = t->t_next) {
@@ -476,9 +485,13 @@ hdr_print(h, fp)
 		putc('\n', fp);
 		break;
 	case DateTime:
-		if (h->h_contents.d > 0L) {
+		/* Use this ONLY if no original lines are available! */
+		if (h->h_contents.d > 0L && h->h_lines == NULL) {
 			cp = (char *)rfc822date(&(h->h_contents.d));
-			fwrite(cp, 1, strlen(cp), fp);
+			s = cp + strlen(cp) -1;
+			if (*cp != 0 && *s == '\n') *s = 0;
+			fputs(cp, fp);
+			putc('\n', fp);
 			break;
 		} /* else fall through */
 	default:
