@@ -31,7 +31,7 @@ STATIC int	sslPointer;		/* index into S/SL table */
 STATIC TableOperation	operation;	/* the current operation */
 
 /* convenience */
-STATIC struct address nullAddr = { NULL, NULL, newAddress, 0, NULL };
+STATIC struct address nullAddr = { NULL, NULL, newAddress, 0, NULL, NULL };
 STATIC struct token atsigntoken = { (u_char *)"@", 1, Special, NULL };
 
 /* Tracing Control */
@@ -99,6 +99,7 @@ STATIC struct receivedtag rt[] = {
 	{	"with",		4,	tWith		},
 	{	"via",		3,	tVia		},
 	{	"for",		3,	tFor		},
+	{	"convert",	7,	tConvert	},
 	{	0,		0,	tSyntaxError	}
 };
 
@@ -157,7 +158,7 @@ SslError(errCode, ap)
 		}
 	}
 	if (t == NULL)
-		t = makeToken((u_char *)"Unknown", strlen("Unknown"));
+		t = makeToken((u_char *)"Unknown", sizeof("Unknown")-1);
 	t->t_type = Error;
 	EmitToken(t, cError, ap);
 
@@ -626,6 +627,27 @@ revaddress(ap)
 	return ap;
 }
 
+void
+revappend(tprev,ap)
+struct token **tprev;
+struct address *ap;
+{
+	struct addr *p;
+	struct token *t;
+
+	ap = revaddress(ap);
+	/* append all the tokens together */
+	if (ap == NULL || ap->a_tokens == NULL
+	    || ap->a_stamp == BadAddress)
+		return;
+	for (p = ap->a_tokens; p ; p = p->p_next) {
+		*tprev = p->p_tokens;
+		for (t = p->p_tokens; t ; t = t->t_next)
+			tprev = &(t->t_next);
+	}
+	*tprev = NULL;
+}
+
 /*
  * The Parser.
  *
@@ -652,7 +674,7 @@ parse822(entry, tlistp, ltmp, tfp)
 	FILE *tfp;
 {
 	int		inReceived, i;
-	struct token	*t, **tprev, *torig;
+	struct token	*t, *torig;
 	struct addr	*p;
 	struct address	*ap, *na, *pap;
 	struct received	rcvd;
@@ -911,7 +933,7 @@ parse822(entry, tlistp, ltmp, tfp)
 		case oEnterReceived:
 			inReceived = 1;
 			rcvd.r_from = rcvd.r_by = rcvd.r_id = rcvd.r_for = NULL;
-			rcvd.r_via = rcvd.r_with = NULL;
+			rcvd.r_via = rcvd.r_with = rcvd.r_convert = NULL;
 			rcvd.r_time = 0L;
 			break;
 		case oExitReceived:
@@ -936,18 +958,10 @@ parse822(entry, tlistp, ltmp, tfp)
 				}
 				break;
 			case rcWith:
-				ap = revaddress(ap);
-				/* append all the tokens together */
-				if (ap == NULL || ap->a_tokens == NULL
-				    || ap->a_stamp == BadAddress)
-					break;
-				tprev = &rcvd.r_with;
-				for (p = ap->a_tokens; p ; p = p->p_next) {
-					*tprev = p->p_tokens;
-					for (t = p->p_tokens; t ; t = t->t_next)
-						tprev = &(t->t_next);
-				}
-				*tprev = NULL;
+				revappend(&rcvd.r_with,ap);
+				break;
+			case rcConvert:
+				revappend(&rcvd.r_convert,ap);
 				break;
 			case rcId:
 				rcvd.r_id = revaddress(ap);
