@@ -230,7 +230,7 @@ main(argc, argv)
 		/* X: check if another daemon is running already */
 		if (!verbose
 		    && (offout < ftell(stdout) || offerr < ftell(stderr))) {
-			fprintf(stderr, "%d %d %d %d\n", offout, ftell(stdout),
+			fprintf(stderr, "%ld %ld %ld %ld\n", offout, ftell(stdout),
 				offerr, ftell(stderr));
 			fprintf(stderr, "%s: daemon not started.\n", progname);
 			die(1, "too many scheduler daemons");
@@ -453,10 +453,13 @@ schedule(fd, file, cehead)
 	  unctlfile(cfp);
 	  return NULL;
 	}
-	if (cfp->contents != NULL && cfp->vfp == NULL) {
-	  free(cfp->contents);
-	  cfp->contents = NULL;
-	}
+	/* -- DO NOT FREE CFP->CONTENTS YET! 
+
+	   if (cfp->contents != NULL && cfp->vfp == NULL) {
+	   free(cfp->contents);
+	   cfp->contents = NULL;
+	   }
+	*/
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
 	  if (subsubdirs) {	/* channel/host/file */
 	    dir = vtx_to_dir(vp, buf);
@@ -472,7 +475,7 @@ schedule(fd, file, cehead)
 	    if (cfp->vfp != NULL)
 	      fprintf(cfp->vfp, "linked to %s\n", path);
 	    if (verbose)
-	      printf("linked %s,%d to %s\n",
+	      printf("linked %s,%ld to %s\n",
 		     vp->cfp->mid, vp->cfp->id, path);
 	  }
 
@@ -492,7 +495,6 @@ slurp(fd)
 	char *contents;
 	long *offset, *ip, *lp;
 	int offsetspace;
-	int msgbodyoffset = -1;
 	struct stat stbuf;
 	struct ctlfile *cfp;
 
@@ -513,6 +515,20 @@ slurp(fd)
 	}
 	contents[stbuf.st_size] = 0;
 
+	cfp = (struct ctlfile *)emalloc((u_int) (sizeof(struct ctlfile)));
+	bzero((void*)cfp,(u_int) (sizeof(struct ctlfile)));
+
+	cfp->fd = fd;
+	cfp->vfp = NULL;
+	cfp->uid = stbuf.st_uid;
+	cfp->mark = V_NONE;
+	cfp->haderror = 0;
+	cfp->contents = contents;
+	cfp->envid    = NULL;
+	cfp->logident = NULL;
+	cfp->erroraddr = NULL;
+	cfp->msgbodyoffset = 0;
+
 	/* go through the file and mark it off */
 	i = 0;
 	offsetspace = 100;
@@ -527,9 +543,7 @@ slurp(fd)
 		offset = (long*)erealloc(offset,sizeof(long)*offsetspace);
 	      }
 	      offset[i++] = s - contents;
-	      if (*s == _CF_BODYOFFSET) {
-		msgbodyoffset = atoi(s+2);
-	      } else if (*s == _CF_MSGHEADERS) {
+	      if (*s == _CF_MSGHEADERS) {
 		/* find a \n\n combination */
 		while (!(*s == '\n' && *(s+1) == '\n'))
 		  if (s-contents < stbuf.st_size)
@@ -539,25 +553,24 @@ slurp(fd)
 		if (s - contents >= stbuf.st_size) {
 		  /* XX: header ran off file */
 		}
-	      }
+	      } else if (*s == _CF_BODYOFFSET)
+		cfp->msgbodyoffset = atoi(s+2);
+	      else if (*s == _CF_DSNENVID)
+		cfp->envid = s+2;
+	      else if (*s == _CF_MESSAGEID)
+		cfp->mid = s+2;
+	      else if (*s == _CF_LOGIDENT)
+		cfp->logident = s+2;
+	      else if (*s == _CF_ERRORADDR)
+		cfp->erroraddr = s+2;
 	    }
 	  }
 	}
-	cfp = (struct ctlfile *)emalloc((u_int) (sizeof(struct ctlfile) +
-						 i * (sizeof offset[0])));
-	bzero((void*)cfp,(u_int) (sizeof(struct ctlfile) +
-				  i * (sizeof offset[0])));
-	/* closing fd must be done in vtxprep(), so we can unlock stuff easy */
-	cfp->fd = fd;
-	cfp->vfp = NULL;
-	cfp->uid = stbuf.st_uid;
 	cfp->nlines = i;
-	cfp->mark = V_NONE;
-	cfp->haderror = 0;
-	cfp->contents = contents;
-	cfp->logident = NULL;
-	cfp->erroraddr = NULL;
-	cfp->msgbodyoffset = msgbodyoffset;
+	/* closing fd must be done in vtxprep(), so we can unlock stuff easy */
+	cfp = (struct ctlfile *)erealloc((void*)cfp,
+					 (u_int) (sizeof(struct ctlfile) +
+						  i * (sizeof offset[0])));
 	lp = &(cfp->offset[0]);
 	ip = &(offset[0]);
 	/* copy over the offsets */
@@ -573,6 +586,7 @@ struct offsort {
 	long	offset;
 	int	myidx;
 	long	headeroffset;
+	long	drptoffset;
 };
 
 /* Used by the qsort comparison routine, bcp points to the control file bytes */
@@ -642,23 +656,82 @@ vtxprep(cfp)
 	      offarr[opcnt].offset = *lp + 2;
 	      offarr[opcnt].myidx = i;
 	      offarr[opcnt].headeroffset = -1;
+	      offarr[opcnt].drptoffset = -1;
 	      ++opcnt;
 	      break;
+	    case _CF_RCPTNOTARY:
+	      /* XX: IETF-NOTARY-DRPT DATA! */
+	      offarr[svn].drptoffset = *lp + 2;
+#if 0 /* ---------------------------------------------------------------- */
+	      /* This fragment is pulled from transports/libta/ctlopen.c */
+	      ++s;
+	      while (*s) {
+		while (*s && (*s == ' ' || *s == '\t')) ++s;
+		if (cistrncmp("RET=",s,4)==0) {
+		  s += 4;
+		  if (cistrncmp("YES",s,3)==0) {
+		    rp->dsnflags |= _DSN_RET_YES;
+		    s += 3;
+		  } else if (cistrncmp("NO",s,2)==0) {
+		    rp->dsnflags |= _DSN_RET_NO;
+		    s += 2;
+		  } else 
+		    /* Not YES/NO ??? */
+		    ;
+		  while (*s && *s != ' ' && *s != '\t') ++s;
+		  continue;
+		}
+		if (cistrncmp("NOTIFY=",s,7)==0) {
+		  s += 7;
+		  if (cistrncmp("SUCCESS",s,7)==0) {
+		    rp->dsnflags |= _DSN_NOTIFY_SUCCESS;
+		    s += 7;
+		  } else if (cistrncmp("FAILURE",s,7)==0) {
+		    rp->dsnflags |= _DSN_NOTIFY_FAILURE;
+		    s += 7;
+		  } else if (cistrncmp("ALWAYS",s,6)==0) {
+		    rp->dsnflags |= _DSN_NOTIFY_ALWAYS;
+		    s += 6;
+		  } else if (cistrncmp("NEVER",s,5)==0) {
+		    rp->dsnflags |= _DSN_NOTIFY_NEVER;
+		    s += 5;
+		  } else
+		    /* None of the above ?? */
+		    ;
+		  while (*s && *s != ' ' && *s != '\t') ++s;
+		  continue;
+		}
+		if (cistrncmp("ORCPT=",s,6)==0) {
+		  s += 6;
+		  rp->orcpt = s;
+		  while (*s && *s != ' ' && *s != '\t') ++s;
+		  if (*s) *s++ = 0;
+		  continue;
+		}
+	      }
+#endif /* ---------------------------------------------------------------- */
+	      break;
 	    case _CF_MSGHEADERS:
 	      for (/* we count up.. */; svn < opcnt; ++svn)
 		offarr[svn].headeroffset = *lp + 2;
 	      break;
+	    case _CF_DSNENVID:
+	      cfp->envid = cp+2;
+	      break;
 	    case _CF_DIAGNOSTIC:
 	      cfp->haderror = 1;
 	      break;
 	    case _CF_MESSAGEID:
-	      cfp->mid = strsave(cp+2);
+	      cfp->mid = cp+2;
+	      break;
+	    case _CF_BODYOFFSET:
+	      cfp->msgbodyoffset = atoi(cp+2);
 	      break;
 	    case _CF_LOGIDENT:
-	      cfp->logident = strsave(cp+2);
+	      cfp->logident = cp+2;
 	      break;
 	    case _CF_ERRORADDR:
-	      cfp->erroraddr = strsave(cp+2);
+	      cfp->erroraddr = cp+2;
 	      break;
 	    case _CF_OBSOLETES:
 	      deletemsg(cp+2, cfp);
@@ -705,10 +778,6 @@ vtxprep(cfp)
 	  spl = sp_lookup((u_int)cfp->id, spt_mesh[L_CTLFILE]);
 	  if (spl != NULL)
 	    sp_delete(spl, spt_mesh[L_CTLFILE]);
-	  if (cfp->mid)
-	    free(cfp->mid);
-	  if (cfp->logident)
-	    free(cfp->logident);
 	  if (cfp->vfp != NULL) {
 	    fprintf(cfp->vfp,
 		    "aborted due to missing information\n");
@@ -789,7 +858,6 @@ vtxprep(cfp)
 	      vp->next[L_CTLFILE] = NULL;
 	      vp->prev[L_CTLFILE] = pvp;
 	      vp->message = NULL;
-	      vp->notary  = NULL;
 	      vp->retryindex = 0;
 	      vp->nextitem = NULL;
 	      vp->previtem = NULL;
@@ -797,7 +865,9 @@ vtxprep(cfp)
 	      vp->proc = NULL;
 	      vp->ngroup = i - svn;
 	      vp->attempts = 0;
-	      vp->headeroffset = offarr[svn].headeroffset; /* Just any of them will do */
+	      vp->notary = NULL;
+	      vp->headeroffset = offarr[i].headeroffset; /* Just any of them will do */
+	      vp->drptoffset = offarr[i].drptoffset;
 	      while (svn < i) {
 		vp->index[i-svn-1] = offarr[svn].myidx;
 		++svn;
@@ -830,7 +900,6 @@ vtxprep(cfp)
 	  vp->next[L_CTLFILE] = NULL;
 	  vp->prev[L_CTLFILE] = pvp;
 	  vp->message = NULL;
-	  vp->notary  = NULL;
 	  vp->retryindex = 0;
 	  vp->nextitem = NULL;
 	  vp->previtem = NULL;
@@ -838,7 +907,9 @@ vtxprep(cfp)
 	  vp->wakeup = 0;
 	  vp->ngroup = i - svn;
 	  vp->attempts = 0;
+	  vp->notary = NULL;
 	  vp->headeroffset = offarr[svn].headeroffset; /* Just any of them will do */
+	  vp->drptoffset = offarr[svn].drptoffset;
 	  while (svn < i) {
 	    vp->index[i-svn-1] = offarr[svn].myidx;
 	    ++svn;
