@@ -9,6 +9,7 @@
 
 char *VerbID = "ZMailer SMTP server %s";
 char *Copyright = "Copyright 1990 Rayan S. Zachariassen";
+char *Copyright2= "Copyright 1991-1994 Matti Aarnio";
 
 /*
  * The smtpserver connects to the router to ask it various questions, like,
@@ -25,6 +26,7 @@ char *Copyright = "Copyright 1990 Rayan S. Zachariassen";
 #define	RKEY_VERIFY	"verify"	/* verify this address */
 #define	RKEY_EXPAND	"expand"	/* expand this address */
 
+#define SMTPLINESIZE	8192
 
 #include <stdio.h>
 #include "hostenv.h"
@@ -119,34 +121,37 @@ Command state;
 struct command {
 	char	*verb;
 	Command	cmd;	
-} command_list[] = {	{	"helo",		Hello		},
-			{	"mail",		Mail		},
-			{	"rcpt",		Recipient	},
-			{	"data",		Data		},
-			{	"send",		Send		},
-			{	"soml",		SendOrMail	},
-			{	"saml",		SendAndMail	},
-			{	"rset",		Reset		},
-			{	"vrfy",		Verify		},
-			{	"expn",		Expand		},
-			{	"help",		Help		},
-			{	"noop",		NoOp		},
-			{	"quit",		Quit		},
-			{	"turn",		Turn		},
-			/* sendmail extensions */
-			{	"onex",		NoOp		},
-			{	"verb",		Verbose		},
-			/* bsmtp extensions */
-			{	"tick",		Tick		},
+} command_list[] = {
 			/* 8-bit smtp extensions */
-			{	"ehlo",		Hello2		},
-			{	"emal",		Mail2		},
-			{	"esnd",		Send2		},
-			{	"esom",		Send2		},
-			{	"esam",		Send2		},
-			{	"evfy",		Verify2		},
+			{	"EHLO",		Hello2		},
+			/* Normal stuff.. */
+			{	"HELO",		Hello		},
+			{	"MAIL",		Mail		},
+			{	"RCPT",		Recipient	},
+			{	"DATA",		Data		},
+			{	"SEND",		Send		},
+			{	"SOML",		SendOrMail	},
+			{	"SAML",		SendAndMail	},
+			{	"RSET",		Reset		},
+			{	"VRFY",		Verify		},
+			{	"EXPN",		Expand		},
+			{	"HELP",		Help		},
+			{	"NOOP",		NoOp		},
+			{	"QUIT",		Quit		},
+			{	"TURN",		Turn		},
+			/* sendmail extensions */
+			{	"ONEX",		NoOp		},
+			{	"VERB",		Verbose		},
+			/* bsmtp extensions */
+			{	"TICK",		Tick		},
+			/* 8-bit smtp extensions */
+			{	"EMAL",		Mail2		},
+			{	"ESND",		Send2		},
+			{	"ESOM",		Send2		},
+			{	"ESAM",		Send2		},
+			{	"EVFY",		Verify2		},
 			/* To fool loosers.. */
-			{	"debug",	DebugMode	},
+			{	"DEBUG",	DebugMode	},
 			{	0,		Null		}
 };
 
@@ -198,6 +203,18 @@ char	*style = "ve";
 #define	WITH_BSMTP	"BSMTP"
 char	*with_protocol = WITH_SMTP;
 
+struct command *carp = NULL;
+FILE *mfp = NULL;
+struct smtpconf *cfinfo = NULL;
+long availspace = 0;	/* available diskspace/2 in bytes	*/
+long sizeoptval = 0;	/* "MAIL FROM:<xxx> SIZE=nnn" -value	*/
+char buf[SMTPLINESIZE];	/* limits size of SMTP commands...
+			   On the other hand, limit is asked
+			   to be only 1000 chars, not 8k.. */
+char helobuf[SMTPLINESIZE];
+
+
+
 unsigned short	auth_flag = 0;
 #ifdef	USE_AUTH
 # ifndef	AUTH_TIMEOUT
@@ -683,28 +700,493 @@ reporterr(tell, msg)
 	}
 }
 
+void
+smtp_helo(cp)
+char *cp;
+{
+	if (state != Hello && state != MailOrHello) {
+		switch (state) {
+		case Mail:
+			cp = "Waiting for MAIL command";
+			break;
+		case Recipient:
+			cp = "Waiting for RCPT command";
+			break;
+		default:
+			cp = NULL;
+			break;
+		}
+		type(503, cp);
+		return;
+	}
+	while (*cp == ' ' || *cp == '\t') ++cp;
+	if (sscanf(cp, "%s", helobuf) != 1) {
+		type(501, "What is your domain name?");
+		return;
+	}
+	/*
+	 * Craig P. says we have to spit back syntactically
+	 * invalid helo parameters at this stage, which is
+	 * hard to do right since it requires a full '822
+	 * tokenizer.  We do a half-hearted attempt here.
+	 */
+	/*
+	 * Matti A. says we have a more-than-half-hearted
+	 * tokenizer -- RFC821SCN.C, lets use it :)
+	 * We need it for proper handling of ESMTP anyway
+	 */
+	if (partridge(helobuf, cfinfo)) {
+		rfc821_error_ptr += (cp - helobuf);
+		type821err(501, buf,
+			   "Invalid `%s' parameter!  Err: %s",
+			   buf, rfc821_error);
+		return;
+	}
+	cfinfo = findcf(helobuf);
+	if (cfinfo != NULL && *(cfinfo->flags) == '!') {
+	  if (cfinfo->flags[1] != '\0')
+	    type(501, "%s", (cfinfo->flags)+1);
+	  else
+	    type(501, "Sorry, access denied.");
+	  return;
+	}
+	/* Check helobuf corresponds to the reverse address */
+	if (ihostaddr[0] != '\0'
+	    && rhostname[0] != '\0' && rhostname[0] != '['
+	    && cistrcmp(helobuf, rhostname) != 0) {
+		type(-250, " %s That hostname is inconsistent with",
+		     myhostname);
+		type(-250, " %s your address to name mapping.",
+		     myhostname);
+		type(carp->cmd==Hello2 ? -250: 250,
+		     " %s expected \"%s %s\"", myhostname,
+		     carp->verb, rhostname);
+	} else
+	  type(carp->cmd==Hello2 ? -250: 250,
+	       " %s Hello %s", myhostname, helobuf);
+	if (carp->cmd == Hello2) {
+		if (mfp == NULL &&
+		    (mfp = mail_open(MSG_RFC822)) == NULL) {
+			type(452, (char *)NULL);
+			return;
+		}
+
+		/* ESMTP -- RFC 1651,1652,1653,1428 thingies */
+		type(-250,"SIZE");
+		type(-250,"8BITMIME");
+		type(-250,"EXPN");
+		type(-250,"VRFY");
+		type(-250,"XDSN"); /* XX: Change to "DSN" when all parts of it function! */
+		type( 250,"HELP");
+		with_protocol = WITH_ESMTP;
+	}
+	state = MailOrHello;
+}
+
+void
+smtp_mail(cp,insecure)
+char *cp;
+int insecure;
+{
+	char *s;
+	int rc;
+	char *drpt_envid;
+	STATFSTYPE statbuf;	/* SysV and BSD definitions differ..	*/
+
+	/* For ESMTP SIZE-option use we need to know
+	   how much space we have, it is easiest by
+	   opening the spool file here, and asking
+	   filesystem to report free space..
+	   We use at most half of the free space,
+	   and each recipient will use the claimed
+	   size, thus marking up its reception..	*/
+
+	if (mfp == NULL
+	    && (mfp = mail_open(MSG_RFC822)) == NULL) {
+		type(452, (char *)NULL);
+		return;
+	}
+#ifdef	FSTATFS
+	if ((rc = FSTATFS(fileno(mfp), &statbuf)) == 0) {
+	  availspace = statbuf.f_bavail * statbuf.f_frsize;
+	}
+#else
+	if ((rc = fstatfs(fileno(mfp),
+			  &statbuf,sizeof statbuf,0)) == 0){
+	  availspace = statbuf.f_bfree * statbuf.f_bsize;
+	}
+#endif
+	if (availspace < 0)
+	  availspace = 2000000000; /* Over 2G ? */
+	availspace >>= 1;
+
+	if (carp->cmd == Mail2 || carp->cmd == Send2) {
+	  with_protocol = WITH_ESMTP;
+	}
+	if (state != Mail && state != MailOrHello) {
+		switch (state) {
+		case Hello:
+			cp = "Waiting for HELO/EHLO command";
+			break;
+		case Recipient:
+			cp = "Waiting for RCPT command";
+			break;
+		default:
+			cp = NULL;
+			break;
+		}
+		type(503, cp);
+		return;
+	}
+	if (cistrncmp(cp, "From:", 5) != 0) {
+		type(501, "where is From: in that?");
+		return;
+	}
+	for (cp = cp+5; *cp != '\0' && *cp != '<'; ++cp)
+		if (!isascii(*cp) || !isspace(*cp))
+			return;
+	if (*cp == '\0') {
+		type(501, "where is <...> in that?");
+		return;
+	} else if (*cp != '<') {
+		type(501, "strangeness between : and <");
+		return;
+	}
+	if (*(cp+1) == '<') {
+		type(501, "there are too many <'s in that!");
+		return;
+	}
+
+	/* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
+	s = rfc821_path(cp,STYLE(cfinfo, 'R'));
+	if (s == cp) {
+	  /* Failure.. */
+	  type821err(501,buf,"Path data: %s",rfc821_error);
+	  return;
+	}
+	if (*s == '>') {
+		type(501, "there are too many >'s in that!");
+		return;
+	}
+	++cp;		/* Skip the initial '<' */
+	*(s-1) = 0;	/* Scrub the final  '>' */
+
+	/* BODY=8BITMIME SIZE=nnnn ENVID=xxxxxx */
+	sizeoptval = -1;
+	drpt_envid = NULL;
+	rc = 0;
+	while (*s) {
+		while (*s && (*s == ' '||*s == '\t')) ++s;
+		if (cistrncmp("BODY=",s,5)==0) {
+			/* Actually we do not use this data. */
+			s += 5;
+			if (cistrncmp(s,"8BITMIME",8)==0) {
+			  s += 8;
+			} else if (cistrncmp(s,"7BIT",4)==0) {
+			  s += 4;
+			}
+			if (*s && *s != ' ' && *s != '\t') {
+			  type(455,"BODY-param data error, must be either 8BITMIME, or 7BIT");
+			  rc = 1;
+			  break;
+			}
+			continue;
+		}
+		if (cistrncmp("SIZE=",s,5)==0) {
+			s += 5;
+			if (sizeoptval != -1) {
+			  type(455,"SIZE-param double definition!");
+			  rc = 1;
+			  break;
+			}
+			/* This data we use, gather the value*/
+			sizeoptval = 0;
+			while (isascii(*s) && isdigit(*s)) {
+				sizeoptval *= 10;
+				sizeoptval += (*s - '0');
+				++s;
+			}
+			if (*s && *s != ' ' && *s != '\t') {
+			  type(501,"SIZE-param data error");
+			  rc = 1;
+			  break;
+			}
+			continue;
+		}
+		/* IETF-NOTARY  SMTP-DRPT extensions */
+		if (cistrncmp("ENVID=",s,6)==0) {
+			if (drpt_envid != NULL) {
+			  type(501,"ENVID double definition!");
+			  rc = 1;
+			  break;
+			}
+			drpt_envid = s+6;
+			while (*s && *s != ' ' && *s != '\t') {
+			  if (*s < 33 || *s > 126) {
+			    type(501,"ENVID data contains illegal characters!");
+			    rc = 1;
+			    break;
+			  }
+			  ++s;
+			}
+			*s++ = 0;
+			if (strlen(drpt_envid) < 6) {
+			  type(501,"ENVID= missing data!");
+			  rc = 1;
+			  break;
+			}
+			continue;
+		}
+		type(555,"Unknown MAIL FROM:<> parameter: %s",s);
+		break;
+	}
+	if (rc != 0) return; /* Error(s) in previous loop.. */
+
+	/*printf("  <path>: len=%d \"%s\"\n",cp-s,cp);*/
+
+	s = NULL;
+	if (*cp != '\0' && STYLE(cfinfo, 'f')) {
+		if ((s = router(RKEY_FROM, 1, cp)) == NULL)
+			/* the error was printed in router() */
+			return;
+		if (atoi(s) / 100 != 2) {
+			/* verification failed */
+			type(atoi(s), s+4, "Failed", "Failed");
+			free(s);
+			return;
+		}
+	}
+
+	if (mfp == NULL &&
+	    (mfp = mail_open(MSG_RFC822)) == NULL) {
+		if (s)
+			free(s);
+		type(452, (char *)NULL);
+		return;
+	}
+	rewind(mfp);
+	if (insecure)
+		fprintf(mfp, "external\n");
+	if (ihostaddr[0] != '\0')
+		fprintf(mfp, "rcvdfrom %s (%s)\n",
+			helobuf, ihostaddr);
+	else
+		fprintf(mfp, "rcvdfrom %s\n", helobuf);
+	fprintf(mfp, "with %s\n", with_protocol);
+#ifdef USE_AUTH
+	if (auth_flag)
+		fprintf(mfp, "authinfo %s\n", auth_username);
+#endif
+	if (*cp == '\0')
+		fprintf(mfp, "channel error\n");
+	else
+		fprintf(mfp, "from <%s>\n", cp);
+	if (drpt_envid != NULL)
+		fprintf(mfp, "envid %s\n", drpt_envid);
+	if (ferror(mfp)) {
+		type(452, (char *)NULL);
+		clearerr(mfp);
+	} else if (sizeoptval > availspace) {
+		type(452, "Try again later, insufficient storage available at the moment");
+	} else if (s)
+		type(atoi(s), s+4, "Ok", "Ok");
+	else
+		type(250, (char *)NULL);
+	if (s)
+		free(s);
+	state = Recipient;
+}
+
+void
+smtp_rcpt(cp)
+char *cp;
+{
+	char *s;
+	char *drpt_notify, *drpt_ret, *drpt_orcpt;
+
+	if (state != Recipient && state != RecipientOrData) {
+		switch (state) {
+		case Hello:
+			cp = "Waiting for HELO command";
+			break;
+		case Mail:
+		case MailOrHello:
+			cp = "Waiting for MAIL command";
+			break;
+		default:
+			cp = NULL;
+			break;
+		}
+		type(503, cp);
+		return;
+	}
+	if (cistrncmp(cp, "To:", 3) != 0) {
+		type(501, "where is To: in that?");
+		return;
+	}
+	for (cp = cp+3; *cp != '\0' && *cp != '<'; ++cp)
+		if (!isascii(*cp) || !isspace(*cp))
+			return;
+	if (*cp == '\0') {
+		type(501, "where is <...> in that?");
+		return;
+	} else if (*cp != '<') {
+		type(501, "strangeness between : and <");
+		return;
+	}
+	if (*(cp+1) == '<') {
+		type(501, "there are too many <'s in that!");
+		return;
+	}
+
+	/* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
+	s = rfc821_path(cp,STYLE(cfinfo, 'R'));
+	if (s == cp) {
+	  /* Failure.. */
+	  type821err(501,buf,"Path data: %s",rfc821_error);
+	  return;
+	}
+	if (*s == '>') {
+		type(501, "there are too many >'s in that!");
+		return;
+	}
+	++cp;		/* Skip the initial '<' */
+	*(s-1) = 0;	/* Scrub the final  '>' */
+
+	/*printf("  <path>: len=%d \"%s\"\n",cp-s,cp);*/
+
+	if (*cp == '\0') {
+		type(501, "what is a null recipient?");
+		return;
+	}
+
+	drpt_notify = NULL;
+	drpt_ret    = NULL;
+	drpt_orcpt  = NULL;
+
+	while (*s) {
+		while (isascii(*s) && isspace(*s)) ++s;
+		/* IETF-NOTARY  SMTP-RCPT-DRPT extensions */
+		if (cistrncmp("NOTIFY=",s,7)==0) {
+		  if (drpt_notify) {
+		    type(501,"NOTIFY-param double defined!");
+		    return;
+		  }
+		  drpt_notify = s;
+		  s += 7;
+		  if (cistrncmp("SUCCESS",s,7)==0)
+		    s += 7;
+		  else if (cistrncmp("FAILURE",s,7)==0)
+		    s += 7;
+		  else if (cistrncmp("ALWAYS",s,6)==0)
+		    s += 6;
+		  else if (cistrncmp("NEVER",s,5)==0)
+		    s += 5;
+		  if (*s && *s != ' ' && *s != '\t') {
+		    type(455,"NOTIFY-param data error");
+		    return;
+		  }
+		  if (*s) *s++ = '\0';
+		  continue;
+		}
+		if (cistrncmp("RET=",s,4)==0) {
+		  if (drpt_ret) {
+		    type(501,"RET-param double defined!");
+		    return;
+		  }
+		  drpt_ret = s;
+		  s += 4;
+		  if (cistrncmp("YES",s,3)==0)
+		    s += 3;
+		  else if (cistrncmp("NO",s,2)==0)
+		    s += 2;
+		  if (*s && *s != ' ' && *s != '\t') {
+		    type(501,"RET-param data error");
+		    return;
+		  }
+		  if (*s) *s++ = '\0';
+		  continue;
+		}
+		if (cistrncmp("ORCPT=",s,6)==0) {
+		  if (drpt_orcpt) {
+		    type(501,"ORCPT-param double defined!");
+		    return;
+		  }
+		  drpt_orcpt = s;
+		  while (*s && *s != ' ' && *s != '\t')
+		    ++s;
+		  if (*s) *s++ = '\0';
+		  if (strlen(drpt_orcpt) < 7) {
+		    type(501,"ORCPT-param data error!");
+		    return;
+		  }
+		  continue;
+		}
+		type(555,"Unknown RCPT TO:<> parameter: %s",s);
+		return;
+	}
+
+
+	s = NULL;
+	if (STYLE(cfinfo, 't')) {
+		if ((s = router(RKEY_TO, 1, cp)) == NULL)
+			/* the error was printed in router() */
+			return;
+		if (atoi(s) / 100 != 2) {
+			/* verification failed */
+			type(atoi(s), s+4, "Failed", "Failed");
+			free(s);
+			return;
+		}
+	}
+
+	/* IETF-NOTARY DSN data: */
+	if (drpt_orcpt || drpt_notify || drpt_ret) {
+	  fputs("todsn",mfp);
+	  if (drpt_notify) fprintf(mfp, " %s", drpt_notify);
+	  if (drpt_ret)    fprintf(mfp, " %s", drpt_ret);
+	  if (drpt_orcpt)  fprintf(mfp, " %s", drpt_orcpt);
+	  fputs("\n",mfp);
+	}
+	/* Normal "RCPT TO:<>" data: */
+	fprintf(mfp, "to <%s>\n", cp);
+	if (sizeoptval < 0)
+		sizeoptval = 0;
+	availspace -= sizeoptval;
+	if (ferror(mfp)) {
+		type(452, (char *)NULL);
+		clearerr(mfp);
+	} else if (availspace < 0) {
+		type(452, "insufficient storage space, try again later");
+	} else if (s) {
+		type(atoi(s), s+4, "Ok", "Ok");
+	} else {
+		if (sizeoptval)
+			type(250, "Ok; can accomodate %d byte message",sizeoptval);
+		else
+			type(250, (char *)NULL);
+	}
+	if (s)
+		free(s);
+	state = RecipientOrData;
+}
+
+
 void
 smtpserver(insecure)
 	int insecure;
 {
 	char *cp, *s;
-	struct command *carp;
 	time_t now;
 	long tell;
-	int fd, inum;
-	FILE *mfp;
+	int inum;
 	struct stat stbuf;
-	struct smtpconf *cfinfo = NULL;
 	struct sockaddr_in localsock;
 	int localsocksize;
-	long availspace = 0;	/* available diskspace/2 in bytes	*/
-	long sizeoptval = 0;	/* "MAIL FROM:<xxx> SIZE=nnn" -value	*/
 	int rc;
-	STATFSTYPE statbuf;	/* SysV and BSD definitions differ..	*/
-	char buf[8192];		/* limits size of SMTP commands...
-				   On the other hand, limit is asked
-				   to be only 1000 chars, not 8k.. */
-	char helobuf[sizeof buf/sizeof(char)];
+	int c, co;
+
+	char *eobuf;
 	int VerboseCommand = 0;
 
 	openlogfp();
@@ -764,29 +1246,70 @@ smtpserver(insecure)
 		alarm(SMTP_ALARM_IVAL);
 
 		*buf = 0;
-		if (fgets(buf, sizeof buf, stdin) == NULL || *buf == 0) {
-		  break;	/* EOF */
-		}
-		alarm(0);	/* Cancel the alarm */
+		s = buf;
+		eobuf = buf + sizeof(buf) -1;
+		rc = -1;
+		co = -1;
+		/* Our own  fgets() -- gets also NULs, flags illegals.. */
+		while ((c = getc(stdin)) != EOF && s < eobuf) {
+			if (c == '\n') {
+			  /* *s++ = c; */ /* Don't save it! No need */
+			  break;
+			} else if (co == '\r' && rc < 0)
+			  rc = (s - buf); /* Spurious CR on the input.. */
 
-		if (verbose && !daemon)
-			fprintf(stdout,"%s",buf); /* XX: trace.. */
-		if (strchr(buf, '\n') == NULL) {
-		    type(500, "Line too long");
+			if (c == '\0'       && rc < 0) rc = (s - buf);
+			if ((c & 0x80) != 0 && rc < 0) rc = (s - buf);
+			if (c != '\r' && c != '\t' &&
+			    (c < 32 || c == 127) && rc < 0)
+			  rc = (s - buf);
+			*s++ = c;
+		}
+		*s = '\0';
+		eobuf = s; /* Buf end ptr.. */
+		alarm(0);	/* Cancel the alarm */
+		if (c == EOF && s == buf) {
+			/* XX: ???  Uh, Hung up on us ? */
+			if (mfp != NULL)
+			  mail_abort(mfp);
+			mfp = NULL;
+			break;
+		}
+		if (c != '\n') {
+		    type(500, "Line too long/not terminated with CRLF..");
 		    continue;
 		}
+
+		/* Zap the possible trailing  \r */
+		if ((eobuf > buf) && (eobuf[-1] == '\r'))
+		  *--eobuf = '\0';
+
+		/* Chop the trailing spaces */
+		while ((eobuf > buf) && (eobuf[-1] == ' ' ||
+					 eobuf[-1] == '\t'))
+		  *--eobuf = '\0';
+
+		if (rc >= 0) {
+		  rfc821_error_ptr = buf + rc;
+		  type821err(500, buf,
+			     "Illegal input characters: %s",
+			     ((buf[rc]=='\0') ? "NUL on SMTP input" :
+			      ((buf[rc]&0x80) ? "8-bit char on SMTP input":
+			       "Control chars on SMTP input")));
+		  continue;
+		}
+
+		if (verbose && !daemon)
+			fprintf(stdout,"%s\n",buf); /* XX: trace.. */
 		report("%.100s", buf);
-		for (cp = buf + strlen(buf) - 1;
-		     cp > buf && isascii(*cp) && isspace(*cp); --cp)
-			continue;
-		*++cp = '\0';
+
 		if (logfp != NULL) {
 			fprintf(logfp, "%dr\t%s\n", pid, buf);
 			fflush(logfp);
 		}
 		for (cp = buf; isascii(*cp) && isalpha(*cp); ++cp)
 			continue;
-		if (cp >  buf + 6) {
+		if (cp >  buf + 6) { /* "DEBUG" is longest of them.. */
 			type(550, "Syntax error");
 			continue;
 		}
@@ -812,82 +1335,8 @@ smtpserver(insecure)
 			break;
 		case Hello:
 		case Hello2:
-			if (state != Hello && state != MailOrHello) {
-				switch (state) {
-				case Mail:
-					cp = "Waiting for MAIL command";
-					break;
-				case Recipient:
-					cp = "Waiting for RCPT or DATA command";
-					break;
-				default:
-					cp = NULL;
-					break;
-				}
-				type(503, cp);
-				break;
-			}
-			while (*cp == ' ' || *cp == '\t') ++cp;
-			if (sscanf(cp, "%s", helobuf) != 1) {
-				type(501, "What is your domain name?");
-				break;
-			}
-			/*
-			 * Craig P. says we have to spit back syntactically
-			 * invalid helo parameters at this stage, which is
-			 * hard to do right since it requires a full '822
-			 * tokenizer.  We do a half-hearted attempt here.
-			 */
-			/*
-			 * Matti A. says we have a more-than-half-hearted
-			 * tokenizer -- RFC821SCN.C, lets use it :)
-			 * We need it for proper handling of ESMTP anyway
-			 */
-			if (partridge(helobuf, cfinfo)) {
-				rfc821_error_ptr += (cp - helobuf);
-				type821err(501, buf,
-					   "Invalid `%s' parameter!  Err: %s",
-					   buf, rfc821_error);
-				break;
-			}
-			cfinfo = findcf(helobuf);
-			if (cfinfo != NULL && *(cfinfo->flags) == '!') {
-			  if (cfinfo->flags[1] != '\0')
-			    type(501, "%s", (cfinfo->flags)+1);
-			  else
-			    type(501, "Sorry, access denied.");
-			  break;
-			}
-			/* Check helobuf corresponds to the reverse address */
-			if (ihostaddr[0] != '\0'
-			    && rhostname[0] != '\0' && rhostname[0] != '['
-			    && cistrcmp(helobuf, rhostname) != 0) {
-				type(-250, " %s That hostname is inconsistent with",
-				     myhostname);
-				type(-250, " %s your address to name mapping.",
-				     myhostname);
-				type(carp->cmd==Hello2 ? -250: 250,
-				     " %s expected \"%s %s\"", myhostname,
-				     carp->verb, rhostname);
-			} else
-			  type(carp->cmd==Hello2 ? -250: 250,
-			       " %s Hello %s", myhostname, helobuf);
-			if (carp->cmd == Hello2) {
-				if (mfp == NULL &&
-				    (mfp = mail_open(MSG_RFC822)) == NULL) {
-					type(452, (char *)NULL);
-					break;
-				}
-
-				/* ESMTP -- RFC 1425,1426,1427,1428 thingies */
-				type(-250,"SIZE");
-				type(-250,"8BITMIME");
-				type(-250,"EXPN");
-				type(-250,"VRFY");
-				type( 250,"HELP");
-				with_protocol = WITH_ESMTP;
-			}
-			state = MailOrHello;
+			/* This code is LONG.. */
+			smtp_helo(cp);
 			break;
 		case Mail:
 		case Mail2:
@@ -895,248 +1344,12 @@ smtpserver(insecure)
 		case Send2:
 		case SendOrMail:
 		case SendAndMail:
-			/* For ESMTP SIZE-option use we need to know
-			   how much space we have, it is easiest by
-			   opening the spool file here, and asking
-			   filesystem to report free space..
-			   We use at most half of the free space,
-			   and each recipient will use the claimed
-			   size, thus marking up its reception..	*/
-			if (mfp == NULL
-			    && (mfp = mail_open(MSG_RFC822)) == NULL) {
-				if (s)
-					free(s);
-				type(452, (char *)NULL);
-				break;
-			}
-#ifdef	FSTATFS
-			if ((rc = FSTATFS(fileno(mfp), &statbuf)) == 0) {
-			  availspace = statbuf.f_bavail * statbuf.f_frsize;
-			}
-#else
-			if ((rc = fstatfs(fileno(mfp),
-					  &statbuf,sizeof statbuf,0)) == 0){
-			  availspace = statbuf.f_bfree * statbuf.f_bsize;
-			}
-#endif
-			if (availspace < 0)
-			  availspace = 2000000000; /* Over 2G ? */
-			availspace >>= 1;
-
-			if (carp->cmd == Mail2 || carp->cmd == Send2) {
-			  with_protocol = WITH_ESMTP;
-			}
-			if (state != Mail && state != MailOrHello) {
-				switch (state) {
-				case Hello:
-					cp = "Waiting for HELO/EHLO command";
-					break;
-				case Recipient:
-					cp = "Waiting for RCPT or DATA command";
-					break;
-				default:
-					cp = NULL;
-					break;
-				}
-				type(503, cp);
-				break;
-			}
-			if (cistrncmp(cp, "From:", 5) != 0) {
-				type(501, "where is From: in that?");
-				break;
-			}
-			for (cp = cp+5; *cp != '\0' && *cp != '<'; ++cp)
-				if (!isascii(*cp) || !isspace(*cp))
-					break;
-			if (*cp == '\0') {
-				type(501, "where is <...> in that?");
-				break;
-			} else if (*cp != '<') {
-				type(501, "strangeness between : and <");
-				break;
-			}
-			if (*(cp+1) == '<') {
-				type(501, "there are too many <'s in that!");
-				break;
-			}
-
-			/* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
-			s = rfc821_path(cp,STYLE(cfinfo, 'R'));
-			if (s == cp) {
-			  /* Failure.. */
-			  type821err(501,buf,"Path data: %s",rfc821_error);
-			  break;
-			}
-			if (*s == '>') {
-				type(501, "there are too many >'s in that!");
-				break;
-			}
-			++cp;		/* Skip the initial '<' */
-			*(s-1) = 0;	/* Scrub the final  '>' */
-
-			/* BODY=8BITMIME SIZE=nnnn */
-			sizeoptval = 0;
-			while (*s) {
-				while (isascii(*s) && isspace(*s)) ++s;
-				if (strncmp("BODY=",s,5)==0) {
-					/* Actually we do not use this data. */
-					s += 5;
-					while (isascii(*s) && isalnum(*s)) ++s;
-					continue;
-				}
-				if (strncmp("SIZE=",s,5)==0) {
-					s += 5;
-					/* This data we use, gather the value*/
-					while (isascii(*s) && isdigit(*s)) {
-						sizeoptval *= 10;
-						sizeoptval += (*s - '0');
-						++s;
-					}
-					continue;
-				}
-				type(501,"Unrecognized optional parameters: %s",s);
-			}
-
-			/*printf("  <path>: len=%d \"%s\"\n",cp-s,cp);*/
-
-			s = NULL;
-			if (*cp != '\0' && STYLE(cfinfo, 'f')) {
-				if ((s = router(RKEY_FROM, 1, cp)) == NULL)
-					/* the error was printed in router() */
-					break;
-				if (atoi(s) / 100 != 2) {
-					/* verification failed */
-					type(atoi(s), s+4, "Failed", "Failed");
-					free(s);
-					break;
-				}
-			}
-
-			if (mfp == NULL &&
-			    (mfp = mail_open(MSG_RFC822)) == NULL) {
-				if (s)
-					free(s);
-				type(452, (char *)NULL);
-				break;
-			}
-			rewind(mfp);
-			if (insecure)
-				fprintf(mfp, "external\n");
-			if (ihostaddr[0] != '\0')
-				fprintf(mfp, "rcvdfrom %s (%s)\n",
-					helobuf, ihostaddr);
-			else
-				fprintf(mfp, "rcvdfrom %s\n", helobuf);
-			fprintf(mfp, "with %s\n", with_protocol);
-#ifdef USE_AUTH
-			if (auth_flag)
-				fprintf(mfp, "authinfo %s\n", auth_username);
-#endif
-			if (*cp == '\0')
-				fprintf(mfp, "channel error\n");
-			else
-				fprintf(mfp, "from <%s>\n", cp);
-			if (ferror(mfp)) {
-				type(452, (char *)NULL);
-				clearerr(mfp);
-			} else if (sizeoptval > availspace) {
-				type(452, "Try again later, insufficient storage available at the moment");
-			} else if (s)
-				type(atoi(s), s+4, "Ok", "Ok");
-			else
-				type(250, (char *)NULL);
-			if (s)
-				free(s);
-			state = Recipient;
+			/* This code is LONG.. */
+			smtp_mail(cp,insecure);
 			break;
 		case Recipient:
-			if (state != Recipient && state != RecipientOrData) {
-				switch (state) {
-				case Hello:
-					cp = "Waiting for HELO command";
-					break;
-				case Mail:
-				case MailOrHello:
-					cp = "Waiting for MAIL command";
-					break;
-				default:
-					cp = NULL;
-					break;
-				}
-				type(503, cp);
-				break;
-			}
-			if (cistrncmp(cp, "To:", 3) != 0) {
-				type(501, "where is To: in that?");
-				break;
-			}
-			for (cp = cp+3; *cp != '\0' && *cp != '<'; ++cp)
-				if (!isascii(*cp) || !isspace(*cp))
-					break;
-			if (*cp == '\0') {
-				type(501, "where is <...> in that?");
-				break;
-			} else if (*cp != '<') {
-				type(501, "strangeness between : and <");
-				break;
-			}
-			if (*(cp+1) == '<') {
-				type(501, "there are too many <'s in that!");
-				break;
-			}
-
-			/* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
-			s = rfc821_path(cp,STYLE(cfinfo, 'R'));
-			if (s == cp) {
-			  /* Failure.. */
-			  type821err(501,buf,"Path data: %s",rfc821_error);
-			  break;
-			}
-			if (*s == '>') {
-				type(501, "there are too many >'s in that!");
-				break;
-			}
-			++cp;		/* Skip the initial '<' */
-			*(s-1) = 0;	/* Scrub the final  '>' */
-
-			/*printf("  <path>: len=%d \"%s\"\n",cp-s,cp);*/
-
-			if (*cp == '\0') {
-				type(501, "what is a null recipient?");
-				break;
-			}
-
-			s = NULL;
-			if (STYLE(cfinfo, 't')) {
-				if ((s = router(RKEY_TO, 1, cp)) == NULL)
-					/* the error was printed in router() */
-					break;
-				if (atoi(s) / 100 != 2) {
-					/* verification failed */
-					type(atoi(s), s+4, "Failed", "Failed");
-					free(s);
-					break;
-				}
-			}
-
-			fprintf(mfp, "to <%s>\n", cp);
-			availspace -= sizeoptval;
-			if (ferror(mfp)) {
-				type(452, (char *)NULL);
-				clearerr(mfp);
-			} else if (availspace < 0) {
-				type(452, "insufficient storage space, try again later");
-			} else if (s) {
-				type(atoi(s), s+4, "Ok", "Ok");
-			} else {
-				if (sizeoptval)
-					type(250, "Ok; can accomodate %d byte message",sizeoptval);
-				else
-					type(250, (char *)NULL);
-			}
-			if (s)
-				free(s);
-			state = RecipientOrData;
+			/* This code is LONG.. */
+			smtp_rcpt(cp);
 			break;
 		case Data:
 			if (state != RecipientOrData) {
@@ -1190,7 +1403,7 @@ smtpserver(insecure)
 #endif	/* LOG_INFO */
 				if (logfp != NULL) {
 					fprintf(logfp,
-						       "%d#\t%d: %d bytes\n",
+						       "%d#\t%d: %ld bytes\n",
 						       pid, inum, tell);
 					fflush(logfp);
 				}
@@ -1238,7 +1451,8 @@ smtpserver(insecure)
 			break;
 		case Verbose:
 			type(-250, VerbID, Version);
-			type(250, Copyright);
+			type(-250, Copyright);
+			type(250, Copyright2);
 			VerboseCommand = 1;
 			break;
 		case DebugMode:
@@ -1286,14 +1500,17 @@ help(cfinfo, query)
 	}
 	switch (carp->cmd) {
 	case Hello:
-		TYPE_("helo your.domain.name");
+	case Hello2:
+		TYPE_("EHLO your.domain.name");
+		TYPE_("HELO your.domain.name");
+		TYPE_("\tThe 'EHLO' is for Extended SMTP feature recognition, and is preferred!.");
 		TYPE_("\tIt is polite to introduce yourself before talking.");
 		TYPE ("\tI will in fact ignore you until you do!");
 		break;
 	case Mail:
 	case Mail2:
-		TYPE_("mail from:<sender>");
-		TYPE_("emal from:<sender>");
+		TYPE_("MAIL FROM:<sender> (ESMTP parameters)");
+		TYPE_("EMAL FROM:<sender>");
 		TYPE_("\tSpecify the originator address for the next message.");
 		if (STYLE(cfinfo, 'f')) {
 			TYPE ("\tThe address will be checked before it is accepted.");
@@ -1302,7 +1519,7 @@ help(cfinfo, query)
 		}
 		break;
 	case Recipient:
-		TYPE_("rcpt to:<recipient> (ESMTP parameters)");
+		TYPE_("RCPT TO:<recipient> (ESMTP parameters)");
 		TYPE_("\tSpecify a destination address for the next message.");
 		if (STYLE(cfinfo, 't')) {
 			TYPE ("\tThe address will be checked before it is accepted.");
@@ -1311,30 +1528,30 @@ help(cfinfo, query)
 		}
 		break;
 	case Data:
-		TYPE_("data");
+		TYPE_("DATA");
 		TYPE_("\tStart collecting the message itself.  The text data");
 		TYPE ("\tis terminated by a <CRLF>.<CRLF> combination.");
 		break;
 	case Reset:
-		TYPE_("rset");
+		TYPE_("RSET");
 		TYPE_("\tReset the state of the SMTP server to be ready for");
 		TYPE_("\tthe next message, and abort any current transaction.");
 		TYPE_("");
 		switch (state) {
 		case Hello:
-			cp = "Waiting for \"helo\" command";
+			cp = "Waiting for \"HELO\" command";
 			break;
 		case Mail:
-			cp = "Waiting for \"mail\" command";
+			cp = "Waiting for \"MAIL\" command";
 			break;
 		case MailOrHello:
-			cp = "Waiting for \"mail\" or \"helo\" command";
+			cp = "Waiting for \"MAIL\" or \"EHLO\"/\"HELO\" command";
 			break;
 		case Recipient:
-			cp = "Waiting for \"rcpt\" command";
+			cp = "Waiting for \"RCPT\" command";
 			break;
 		case RecipientOrData:
-			cp = "Waiting for \"rcpt\" or \"data\" command";
+			cp = "Waiting for \"RCPT\" or \"DATA\" command";
 		default:
 			cp = "Unknown";
 			break;
@@ -1351,8 +1568,8 @@ help(cfinfo, query)
 		break;
 	case Verify:
 	case Verify2:
-		TYPE_("vrfy <recipient>");
-		TYPE_("evfy <recipient>");
+		TYPE_("VRFY <recipient>");
+		TYPE_("EVFY <recipient>");
 		if (STYLE(cfinfo, 'v')) {
 			TYPE_("\tPrints the recipients for the given address.")
 			TYPE ("\tIf the address is local, it is not expanded.");
@@ -1361,7 +1578,7 @@ help(cfinfo, query)
 		}
 		break;
 	case Expand:
-		TYPE_("expn <recipient>");
+		TYPE_("EXPN <recipient>");
 		if (STYLE(cfinfo, 'e')) {
 			TYPE_("\tPrints the recipients for the given address.")
 			TYPE ("\tIf the address is local, it is expanded.");
@@ -1374,26 +1591,27 @@ help(cfinfo, query)
 		TYPE ("\tThis command does nothing.");
 		break;
 	case Quit:
-		TYPE_("quit");
+		TYPE_("QUIT");
 		TYPE ("\tTerminate the SMTP protocol conversation.");
 		break;
 	case Verbose:
-		TYPE_("verb");
+		TYPE_("VERB");
 		TYPE_("\tPrints out the SMTP server version and copyright notice.");
 		TYPE ("\tThis command has no other effect.");
 		break;
 	case Tick:
-		TYPE_("tick id");
+		TYPE_("TICK id");
 		TYPE ("\tThis BSMTP command is just reflected back at you.");
 		break;
 	case Help:
-		TYPE_("help [command]");
+		TYPE_("HELP [command]");
 		TYPE_("\tReminder of what the SMTP command does, or prints:");
 		TYPE_("");
 		/* fall through */
 	case Null:
 	default:
 		TYPE_(Copyright);
+		TYPE_(Copyright2);
 		TYPE_("");
 		printf("241-The following commands are recognized:");
 		col = 100;
@@ -1408,14 +1626,15 @@ help(cfinfo, query)
 		}
 		printf("\r\n");
 		TYPE_("");
-		TYPE_("The normal sequence is: helo (mail rcpt+ data)+ quit.");
+		TYPE_("The normal sequence is: EHLO/HELO (MAIL RCPT+ DATA)+ QUIT.");
 		TYPE_("");
-		TYPE_("This mailer always accepts 8-bit and binary message data.");
+		TYPE_("This mailer always accepts 8-bit and binary message data");
+		TYPE_("though you are better to use MIME format!");
 		TYPE_("");
 		printf("241-For local information contact: ");
 		printf("postmaster@%s\r\n", myhostname);
 		printf("241 SMTP server comments and bug reports to: ");
-		printf("zmhacks@cs.toronto.edu\r\n");
+		printf("zmhacks@nic.funet.fi\r\n");
 		break;
 	}
 	fflush(stdout);
@@ -1799,6 +2018,7 @@ void
 type821err(code,inbuf,msg,s1,s2,s3)
 	int code;
 	char *msg, *inbuf;
+	void *s1, *s2, *s3; /* XX: Propably not portable.. */
 {
 	char *s = inbuf+4;
 
@@ -1871,7 +2091,9 @@ mvdata(out)
 	idxnum['\r'] = 1;
 	idxnum['\n'] = 2;
 	idxnum['.'] = 3;
-	/* idxnum[EOF] = 4; */
+#if EOF == -1
+	idxnum[EOF] = 4;
+#endif
 	state = 10;
 	endstate = X_;
 	input = stdin;
@@ -1925,7 +2147,7 @@ readcffile(name)
 	char *name;
 {
 	FILE *fp;
-	struct smtpconf scf, *head, *tail;
+	struct smtpconf scf, *head, *tail = NULL;
 	char c, *cp, buf[1024];
 
 	if ((fp = fopen(name, "r")) == NULL)
