@@ -201,7 +201,7 @@ append_header(va_alist)
 	return linecnt;
 }
 
-void
+static void
 output_content_type(rp,ct)
 struct rcpt *rp;
 struct ct_data *ct;
@@ -293,7 +293,7 @@ struct ct_data *ct;
 	}
 }
 
-struct ct_data *
+static struct ct_data *
 parse_content_type(ct_linep)
 char **ct_linep;	/* Could be multiline! */
 {
@@ -446,7 +446,7 @@ char **ct_linep;	/* Could be multiline! */
 	abort();
 }
 
-struct cte_data *
+static struct cte_data *
 parse_content_encoding(cte_linep)
 char **cte_linep;	/* Propably is not a multiline entry.. */
 {
