@@ -129,6 +129,9 @@ struct utimbuf {
  */
 char *maildirs[] = {
 	"/var/mail",
+	"/usr/mail",
+	"/var/spool/mail",
+	"/usr/spool/mail",
 	0 
 };
 
@@ -790,6 +793,7 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 	long eofindex;
 	char buf[2];
 	struct stat st;
+	char *fromuser;
 
 	if ((fp = fdopen(fdmail, "a+")) == NULL) {
 	  notaryreport(NULL,NULL,NULL);
@@ -858,8 +862,10 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 	  convert_qp = 0; /* If malloc failed, clear this too,
 			     if it was not set, it doesn't change.. */
 
+	fromuser = rp->addr->link->user;
+	if (*fromuser == 0) fromuser = "postmaster";
 	if (fprintf(fp, "%s%s %s",
-		    FROM_,rp->addr->link->user,timestring) == EOF
+		    FROM_,fromuser,timestring) == EOF
 	    || writeheaders(rp,fp,"\n",convert_qp,0) < 0) {
 	  notaryreport(rp->addr->link->user, "failed",
 		       "500 (Write to user's mailbox failed)");
