@@ -157,6 +157,7 @@ int D_alloc = 0;		/* Memory usage debug */
 /* Following options can be declared in ESMTP  EHLO responce  */
 #define ESMTP_SIZEOPT  0x0001
 #define ESMTP_8BITMIME 0x0002
+#define ESMTP_DSN      0x0004
 
 int  ehlo_capabilities = 0;
 long ehlo_sizeval      = 0;
@@ -837,6 +838,9 @@ char *buf;
 
 	if (strncmp(buf,"8BITMIME",8)==0) {
 	  ehlo_capabilities |= ESMTP_8BITMIME;
+	} else if (strcmp(buf,"XDSN")==0 ||
+		   strcmp(buf,"DSN")==0) {
+	  ehlo_capabilities |= ESMTP_DSN;
 	} else if (strncmp(buf,"SIZE ",5)==0 ||
 		   strcmp(buf,"SIZE") == 0) {
 	  ehlo_capabilities |= ESMTP_SIZEOPT;
@@ -923,6 +927,7 @@ smtpconn(host, fparr, verboselog)
 	struct hostent *hp, hent;
 	char	hbuf[MAXHOSTNAMELEN+1], *addrs[2];
 	struct mxdata mxh[MAXFORWARDERS];
+	int	mxcnt = 0;
 
 #ifdef	BIND
 	h_errno = 0;
@@ -960,7 +965,7 @@ smtpconn(host, fparr, verboselog)
 	  addrs[0] = (char *)&ipaddr;
 	  addrs[1] = NULL;
 	  hp_setalist(hp, addrs);
-	  retval = makeconn(hp, &smtpfd, verboselog, fparr);
+	  retval = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
 	} else {
 	  hbuf[0] = '\0';
 	  errno = 0;
@@ -974,7 +979,7 @@ smtpconn(host, fparr, verboselog)
 		|| h_errno == TRY_AGAIN
 #endif /* BIND */
 		) {
-	      sprintf(remotemsg,"500 (gethostbyname(%s): try later)",host);
+	      sprintf(remotemsg,"466 (gethostbyname(%s): try later)",host);
 	      notaryreport(NULL,NULL,remotemsg);
 	      if (verboselog)
 		fprintf(verboselog,"%s\n",remotemsg);
@@ -989,7 +994,7 @@ smtpconn(host, fparr, verboselog)
 	    if ((i = getrr(hbuf, sizeof hbuf, (u_short)T_CNAME),2) == 1) {
 	      host = hbuf;
 	    } else if (h_errno == TRY_AGAIN) {
-	      sprintf(remotemsg, "500 (Canonicalize, try later: %s)", host);
+	      sprintf(remotemsg, "466 (Canonicalize, try later: %s)", host);
 	      notaryreport(NULL,NULL,remotemsg);
 	      if (verboselog)
 		fprintf(verboselog,"%s\n",remotemsg);
@@ -998,7 +1003,7 @@ smtpconn(host, fparr, verboselog)
 		 it is possible that MXs or As are available
 		 even though CNAME lookups fail (see .UK) */
 #else  /* !BIND */
-	    sprintf(remotemsg, "500 (canonicalize lookup failed on %s)", host);
+	    sprintf(remotemsg, "466 (canonicalize lookup failed on %s)", host);
 	    notaryreport(NULL,NULL,remotemsg);
 	    if (verboselog)
 	      fprintf(verboselog,"%s\n", remotemsg);
@@ -1014,11 +1019,12 @@ smtpconn(host, fparr, verboselog)
 	   * Look for MX RR's. If none found, use the hostentry in hp.
 	   * Otherwise loop through all the mxhosts doing gethostbyname's.
 	   */
+	  mxcnt = 0;
 	  if (!noMX) {
 	    int rc;
 	    if (verboselog)
 	      fprintf(verboselog," getmxrr(%s)",host);
-	    rc = getmxrr(host, mxh, sizeof mxh/sizeof mxh[0],0);
+	    rc = getmxrr(host, mxh, sizeof mxh/sizeof mxh[0],0,&mxcnt);
 	    if (verboselog)
 	      fprintf(verboselog," rc=%d, mxh[0].host=%s (host=%s)\n",
 		      rc,mxh[0].host,host);
@@ -1045,7 +1051,7 @@ smtpconn(host, fparr, verboselog)
 		  || h_errno == TRY_AGAIN
 #endif /* BIND */
 		  ) {
-		sprintf(remotemsg,"500 (gethostbyname<%s>: try later)",host);
+		sprintf(remotemsg,"466 (gethostbyname<%s>: try later)",host);
 		notaryreport(NULL,NULL,remotemsg);
 		if (verboselog)
 		  fprintf(verboselog,"%s\n",remotemsg);
@@ -1057,15 +1063,17 @@ smtpconn(host, fparr, verboselog)
 			"500 (configuration inconsistency. MX usage forbidden, no address in DNS: '%s', errno=%s)",
 			host,strerror(errno));
 	      else
-	      sprintf(remotemsg,
-		      "500 (nameserver data inconsistency. No MX, no address: '%s', errno=%s)",
-		      host,strerror(errno));
+		sprintf(remotemsg,
+			"500 (nameserver data inconsistency. No MX, no address: '%s', errno=%s)",
+			host,strerror(errno));
 	      notaryreport(NULL,NULL,remotemsg);
 	      if (verboselog)
 		fprintf(verboselog,"%s\n",remotemsg);
-	      return EX_UNAVAILABLE;
+	      /* it was: EX_UNAVAILABLE, but such blocks retrying, thus
+		 current EX_TEMPFAIL, which will cause timeout latter on.. */
+	      return EX_TEMPFAIL;
 	    }
-	    retval = makeconn(hp, &smtpfd, verboselog, fparr);
+	    retval = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
 	  } else {
 	    retval = EX_UNAVAILABLE;
 	    for (i = 0; mxh[i].host != NULL; ++i) {
@@ -1075,7 +1083,7 @@ smtpconn(host, fparr, verboselog)
 		firstmx = i+1;
 		continue;
 	      }
-	      r = makeconn(hp, &smtpfd, verboselog, fparr);
+	      r = makeconn(hp, &smtpfd, verboselog, fparr, mxcnt);
 	      if (r == EX_OK) {
 		while (mxh[i].host != NULL) {
 		  free(mxh[i].host);
@@ -1104,11 +1112,12 @@ smtpconn(host, fparr, verboselog)
 }
 
 int
-makeconn(hp, fdp, verboselog, fparr)
+makeconn(hp, fdp, verboselog, fparr, mxcnt)
 	struct hostent *hp;
 	int *fdp;
 	FILE *fparr[];
 	FILE *verboselog;
+	int mxcnt;
 {
 	int retval;
 
@@ -1744,10 +1753,13 @@ typedef union {
 } querybuf;
 
 int
-getmxrr(host, mx, maxmx, depth)
+getmxrr(host, mx, maxmx, depth, mxcntp)
 	char *host;
 	struct mxdata mx[];
 	int maxmx, depth;
+	int *mxcntp; /* Return the NUMBER of MXes that were found,
+			including those that were considered faulty.
+			Will count all MXes. */
 {
 	HEADER *hp;
 	msgdata *eom, *cp;
@@ -1775,14 +1787,14 @@ getmxrr(host, mx, maxmx, depth)
 	if (n < 0) {
 	  fprintf(stderr, "res_mkquery failed\n");
 	  sprintf(remotemsg,
-		  "500 (Internal: res_mkquery failed on host: %s)",host);
+		  "466 (Internal: res_mkquery failed on host: %s)",host);
 	  notaryreport(NULL,NULL,remotemsg);
 	  return EX_SOFTWARE;
 	}
 	n = res_send((char *)&qbuf, n, (char *)&answer, sizeof answer);
 	if (n < 0) {
 	  sprintf(remotemsg,
-		  "500 (Internal: res_send failed on host: %s)",host);
+		  "466 (Internal: res_send failed on host: %s)",host);
 	  notaryreport(NULL,NULL,remotemsg);
 	  return EX_TEMPFAIL;
 	}
@@ -1872,33 +1884,40 @@ getmxrr(host, mx, maxmx, depth)
 	  if (n < 0)
 	    break;
 	  cp += n;
-	  /* [mea] Canonicalize this target & get A records.. */
-	  hent = gethostbyname(buf);
-	  if (!hent)
+
+	  hent = gethostbyname(buf); /* This resolves CNAME, should it
+					should not in case of MX server */
+	  if (!hent) {
+	    *mxcntp += 1;
 	    continue;		/* Well well.. spurious! */
-	  mx[nmx].host = (msgdata *)emalloc((u_int)(strlen(hent->h_name)+1));
-	  if (mx[nmx].host == NULL) {
-	    fprintf(stderr, "Out of virtual memory!\n");
-	    exit(EX_OSERR);
 	  }
-	  strcpy(mx[nmx].host, hent->h_name);
 	  if (cistrcmp(hent->h_name, myhostname) == 0)
 	    maxpref = mx[nmx].pref;
 	  if (matchmyaddresses(hent))
 	    maxpref = mx[nmx].pref;
+
+	  if ((mx[nmx].host = (msgdata *)emalloc(strlen(buf)+1)) == NULL) {
+	    fprintf(stderr, "Out of virtual memory!\n");
+	    exit(EX_OSERR);
+	  }
+	  strcpy(mx[nmx].host, buf);
+	  /* [mea] Canonicalize this target & get A records.. */
 	  ++nmx;
 	}
 	if (nmx == 0 && realname[0] != '\0' &&
 	    cistrcmp(host,(char*)realname)) {
 	  /* do it recursively for the real name */
-	  return getmxrr((char *)realname, mx, maxmx, depth+1);
+	  return getmxrr((char *)realname, mx, maxmx, depth+1, mxcntp);
 	} else if (nmx == 0) {
 	  /* "give it benefit of doubt" */
 	  mx[0].host = NULL;
 	  mxcount = 0;
 	  return EX_OK;
 	}
-	/* discard MX RRs with a value >= that of localdomain */
+	for (i = 0; i < nmx; ++i) {
+	}
+
+	/* discard MX RRs with a value >= that of  myhost */
 	if (maxpref >= 0) {
 	  for (n = i = 0; n < nmx; ++n) {
 	    if (mx[n].pref >= maxpref) {
@@ -1930,6 +1949,7 @@ getmxrr(host, mx, maxmx, depth)
 	for (i = 0, n = 0; i < nmx; ++i) {
 	  if (mx[i].host == NULL)
 	    continue;
+	  *mxcntp += 1;
 	  if (n < i) {
 	    mx[n] = mx[i];
 	    mx[i].host = NULL;
@@ -2116,6 +2136,7 @@ rightmx(spec_host, addr_host)
 {
 	struct mxdata mxh[MAXFORWARDERS];
 	int	i;
+	int	mxcnt;
 
 	if (cistrcmp(spec_host, addr_host) == 0)
 	  return 1;
@@ -2123,7 +2144,8 @@ rightmx(spec_host, addr_host)
 	  return 0;
 
 	mxh[0].host = NULL;
-	switch (getmxrr(addr_host, mxh, sizeof mxh/sizeof mxh[0], 0)) {
+	mxcnt = 0;
+	switch (getmxrr(addr_host, mxh, sizeof mxh/sizeof mxh[0], 0, &mxcnt)) {
 	case EX_OK:
 	  if (mxh[0].host == NULL)
 	    return cistrcmp(addr_host, remotehost) == 0;
